/*
** Module   : ControlSetEntity.java
** Abstract : server side grouped control widget parent class
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23487 Created initial version.
** 002 SIY 20060209   @24430 Major refactoring.
** 003 GES 20060803   @28371 Some cleanup and refactoring.
** 004 GES 20061103   @30943 Fixed abend due to the use of Arrays.asList()
**                           which returns an unmodifiable list.
** 005 GES 20061108   @31026 Code and doc cleanup and shift away from the character.getValue().
** 006 GES 20061128   @31452 Fixed setListItemPairs() to add in the order specified rather than
**                           reversing the order. This only affects non-sorted lists as the
**                           sorted lists would be forced into the right order later.
** 007 SIY 20061213   @31673 Maintain dynamic pairs attribute.
** 008 SIY 20101122          Inferred generics.
** 009 SIY 20101124          Fixed UnsupportedOperationException.
** 010 SVL 20101214          Added addFirst(character, character).
** 011 ECF 20110426          Repeated fix #004 for 2 more locations.
** 012 SIY 20111101          Updated imports.
** 013 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 014 OM  20130527          Added int64 support fopr entry().
** 015 CA  20131115          Changes related to properly instantiate dynamic resources.
** 016 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 017 SVL 20141102          Added isSort.
** 018 GES 20141124          Added addLast and delete methods. Enhanced addFirst. Added 
**                           list-items attr and a variant of entry().
** 019 IAS 20150123          Implemented DELETE() method
** 020 IAS 20150202          Implemented ADD_LAST() method, fixed ADD-FIRST()
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.format.*;

/**
 * Server side parent class for widgets that group other controls.
 */
public class ControlSetEntity<T extends ControlSetConfig>
extends ControlEntity<T>
{
   /* Function to access ControlSetItem label */
   protected static final Function<ControlSetItem, String> GET_LABEL = 
         new Function<ControlSetItem, String>()
   {
      @Override
      public String apply(ControlSetItem item)
      {
         return item.getLabel().toString().trim();
      }
   };

   /* Function to access ControlSetItem value */
   protected static final Function<ControlSetItem, String> GET_VALUE = 
         new Function<ControlSetItem, String>()
   {
      @Override
      public String apply(ControlSetItem item)
      {
         return item.getValue().toString().trim();
      }
   };

   /**
    * Construct an instance using the given configuration.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    * @param   csc
    *          Source configuration.
    */
   public ControlSetEntity(boolean dynamic, T csc)
   {
      super(dynamic, csc);
   }
   
   /**
    * Implements the ENTRY() widget method.  This method returns items from
    * the set by 1-based index.
    *
    * @param    index
    *           The index of an item.
    *
    * @return   The list item indexed or <code>unknown value</code>.
    */
   @Override
   public character entry(int64 index)
   {
      if (index == null || index.isUnknown())
         return new character();
      
      return entry(index.intValue());
   }

   /**
    * Implements the ENTRY() widget method.  This method returns items from
    * the set by 1-based index.
    *
    * @param    index
    *           The index of an item.
    *
    * @return   The list item indexed or <code>unknown value</code>.
    */
   @Override
   public character entry(int index)
   {
      ControlSetItem[] items = config.items;
      
      if (index < 1 || index > items.length)
         return new character();

      return items[index - 1].getLabel();
   }

   /**
    * Implements the LOOKUP() widget method.  This method returns the 1-based
    * item's index in the set.
    *
    * @param    text
    *           The text of an item.
    *
    * @return   The item's index or <code>unknown value</code>.
    */
   @Override
   public integer lookup(character text)
   {
      if (text.isUnknown())
         return new integer();

      ControlSetItem[] items = config.items;
      
      for (int i = 0; i < items.length; i ++)
      {
         if (items[i].getLabel().compareTo(text) == 0)
             return new integer(i + 1);
      }

      // if list-string is not in the list, LOOKUP returns 0
      return new integer(0);
   }
   
   /**
    * Implements the LOOKUP() widget method.  This method returns the 1-based
    * item's index in the set.
    *
    * @param    text
    *           The text of an item.
    *
    * @return   The item's index or <code>unknown value</code>.
    */
   @Override
   public integer lookup(String text)
   {
      return lookup(new character(text));
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a list of labels
    * to the beginning of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addFirst(character itemList)
   {
      if (itemList == null)
      {
         itemList = new character(); 
      }

      return addFirst(itemList.toStringMessage());
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a list of labels
    * to the beginning of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addFirst(String itemList)
   {
      return add(itemList, new ItemsListExpander()
      {
         @Override
         public void addItems(String[] elem, ControlSetItem[] items, List<ControlSetItem> list)
         {
            for (String s : elem)
            {
               list.add(controlSetItem(s));
            }
            list.addAll(Arrays.asList(items));
         }
      });
   }

   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(character label, BaseDataType value)
   {
      return add(label, value, new ItemsListByOneExpander()
      {
         @Override
         public ControlSetItem[] addItem(ControlSetItem[] items, character label, 
               BaseDataType value)
         {
            ControlSetItem[] nitems = new ControlSetItem[items.length + 1];
            nitems[0] = controlSetItem(label, value);
            System.arraycopy(items, 0, nitems, 1, items.length);
            return nitems;
         }
      });
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(String label, BaseDataType value)
   {
      return addFirst(new character(label), value);
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character itemList)
   {
      if (itemList == null)
      {
         itemList = new character(); 
      }
      
      return addLast(itemList.toStringMessage());
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String itemList)
   {
      return add(itemList, new ItemsListExpander()
      {
         @Override
         public void addItems(String[] elem, ControlSetItem[] items, List<ControlSetItem> list)
         {
            list.addAll(Arrays.asList(items));
            for (String s : elem)
            {
               list.add(controlSetItem(s));
            }
         }
      });
   }
   
   protected logical add(String itemList, ItemsListExpander expander)
   {
      if(itemList == null) 
      {
         itemList = new character().toStringMessage();
      }
      if(config.pairs) {
         ErrorManager.recordOrShowError(4064, String.format(
               "Missing argument when setting attribute ADD-LAST for %s %s", type(),
               frame.getName(this)), false);
         ErrorManager.recordOrShowError(4065, String.format(
               "The ADD-LAST attribute on the %s %s has invalid arguments", type(),
               frame.getName(this)), false);
         return new logical(false);
      }
      
      String[] elem = itemList.split(String.valueOf(config.delimiter));
      ControlSetItem[] items = config.items;
      if (items == null)
         items = new ControlSetItem[0];
      
      List<ControlSetItem> list = new ArrayList<>(items.length + elem.length);
      expander.addItems(elem, items, list);
      
      setItems(list.toArray(new ControlSetItem[list.size()]));
      return new logical(true);
   }

   private static interface ItemsListExpander {
      public void addItems(String[] elem, ControlSetItem[] items, List<ControlSetItem> list);
   }
   
   private ControlSetItem controlSetItem(String s) {
      return controlSetItem(new character("?".equals(s.trim()) ? "" : s), new character(s)); 
   }

   protected ControlSetItem controlSetItem(character label, BaseDataType value) {
      return new ControlSetItem(config.nextItemId(), label.isUnknown() ? new character(" ") : (character)label.duplicate(), 
            value.duplicate()); 
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, BaseDataType value)
   {
      return add(label, value, new ItemsListByOneExpander()
      {
         @Override
         public ControlSetItem[] addItem(ControlSetItem[] items, character label, 
               BaseDataType value)
         {
            ControlSetItem[] nitems = Arrays.copyOf(items, items.length + 1);
            nitems[items.length] = controlSetItem(label, value);
            return nitems;
         }
      });
   }

   public logical add(character label, BaseDataType value, ItemsListByOneExpander expander)
   {
      if(!config.pairs) {
         ErrorManager.recordOrShowError(8950, String.format(
               "Ignoring extra argument(s) of ADD-LAST for %s %s that is of LIST-ITEMS type", 
               type(), frame.getName(this)), false);
         return addLast(label);
      }
      ControlSetItem[] items = config.items;
      if (items == null)
         items = new ControlSetItem[0];
      setItems(expander.addItem(items, label, value));
      return new logical(true);
   }

   private static interface ItemsListByOneExpander {
      public ControlSetItem[] addItem(ControlSetItem[] items, character label, BaseDataType value);
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String label, BaseDataType value)
   {
      return addLast(new character(label), value);
   }
   
   /**
    * Helper method for the DELETE() widget method, which removes the list item
    * or items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * @param checkAll
    *           return TRUE only if all items where found
    * 
    * @return <code>true</code> if the method succeeds.
    */
   protected logical delete(character itemList, boolean checkAll)
   {
      if (itemList == null ) {
         itemList = new character();
      }
      return delete(itemList.toStringMessage(), checkAll);
   }

   /**
    * Helper method for the DELETE() widget method, which removes the list item
    * or items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * @param checkAll
    *           return TRUE only if all items where found
    * 
    * @return <code>true</code> if the method succeeds.
    */
   protected logical delete(String itemList, boolean checkAll)
   {
      if (itemList == null) {
         itemList = "?";
      }
      return delete(
            new ArrayList<String>(Arrays.asList(itemList.split(new String(
                  new char[] { config.delimiter })))), GET_VALUE, checkAll);
   }

   /**
    * Helper method for the DELETE() widget method, which removes the list item
    * or items specified.
    * 
    * @param itemList
    *           The List of items to delete.
    * @param checkAll
    *           return TRUE only if all items where found
    * 
    * @return <code>true</code> if the method succeeds.
    */
   protected logical delete(List<String> itemList, Function<ControlSetItem, String> extractValue, boolean checkAll)
   {
      boolean foundLast = false;
      ControlSetItem[] items = config.items;
      List<ControlSetItem> nitems = new ArrayList<>(items.length);
      for (ControlSetItem item : items)
      {
         int ip = itemList.indexOf(extractValue.apply(item));
         if (ip < 0)
         {
            nitems.add(item);
            continue;
         }
         if (ip == itemList.size() - 1)
         {
            foundLast = true;
         }
         itemList.remove(ip);
      }
      setItems(nitems.toArray(new ControlSetItem[nitems.size()]));
      return new logical(checkAll ? itemList.isEmpty() : foundLast);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(int64 idx)
   {
      if (idx == null || idx.isUnknown())
         return new logical();

      return delete(idx.longValue());
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(long idx)
   {
      ControlSetItem[] items = config.items;

      if (idx < 0 || idx > items.length)
         return new logical(false);
      if (idx == 0)
      {
         ErrorManager.recordOrShowError(4065, String.format(
               "The DELETE attribute on the %s %s has invalid arguments", type(),
               frame.getName(this)), false);
         return new logical(false);
      }
      idx--;
      ControlSetItem[] nitems = new ControlSetItem[items.length - 1];
      int nidx = 0;
      for (int i = 0; i < items.length; i++)
      {
         if (i == (int) idx)
            continue;
         nitems[nidx++] = items[i];
      }
      setItems(nitems);
      return new logical(true);
   }
   
   /**
    * Store array of items into config.
    * 
    * @param    items
    *           Array of new items.
    */
   public void setItems(ControlSetItem[] items)
   {
      if (config.items != null && items == null)
         return;
      for(ControlSetItem item: items)
      {
         if(item.getId() == 0)
         {
            item.setId(config.nextItemId());
         }
      }
      config.items = items;
      if (!config.realized || config.autoResize)
      {
         config.setHeightChars(items == null ? 0 : items.length);
      }
      
      sortItems();
      
      pushScreenDefinition();
   }
 
   /**
    * Set SORT attribute.
    * 
    * @param    sort
    *           New value for the attribute.
    */
   public void setSort(logical sort)
   {
      setSort(sort.booleanValue());
   }
   
   /**
    * Set SORT attribute.
    * 
    * @param    sort
    *           New value for the attribute.
    */
   public void setSort(boolean sort)
   {
      config.sort = sort;
      sortItems();
      pushScreenDefinition();
   }

   /**
    * Get SORT attribute.
    *
    * @return  value for the attribute.
    */
   public logical isSort()
   {
      return new logical(config.sort);
   }
   
   /**
    * Sets the LIST-ITEMS writeable attribute.
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItems(character list)
   {
      if (list == null || list.isUnknown())
         return;

      setItems(new ControlSetItem[0]);
      
      addFirst(list);
   }
   
   /**
    * Setter for the LIST-ITEMS writeable attribute.
    *
    * @param    list
    *           The new value for the LIST-ITEMS.
    */
   public void setListItems(String list)
   {
      setListItems(new character(list));
   }
   
   /**
    * Common implementation list item pair attributes (LIST-ITEM-PAIRS and
    * RADIO-BUTTONS) attribute. The parameter contains string representation
    * of all labels and values separated by a delimiter (comma is the default
    * delimiter).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,...,LabelN,valueN</code>
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItemPairs(character list)
   {
      if (list.isUnknown())
         return;
      
      String itemList = list.toStringMessage();
      char[] delim = new char[1];
      delim[0] = config.delimiter;
      String d = new String(delim);
      String type = config.dataType;
      String[] elem = itemList.split(d);
      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];

      List<ControlSetItem> newlist = asList(items);
      
      for (int i = 0; i < elem.length; i += 2)
      {
         BaseDataType value = null;
         
         if ((i + 1) < elem.length)
            value = DisplayFormat.instanceOfType(type, elem[i + 1]);
         else
            value = DisplayFormat.instanceOfType(type);
         
         newlist.add(new ControlSetItem(config.nextItemId(), new character(elem[i]), value));
      }
      
      items = newlist.toArray(items);
      
      config.pairs = true;
      setItems(items);
      pushScreenDefinition();
   }
   
   /**
    * Gets the LIST-ITEMS writeable attribute.
    * 
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItems()
   {
      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];
      
      StringBuffer sb = new StringBuffer();
      
      for (int i = 0; i < items.length; i++)
      {
         sb.append(items[i].getLabel().toStringMessage());
         
         if (i < items.length - 1)
            sb.append(config.delimiter);
      }
      
      return new character(sb.toString());
   }

   /**
    * Gets the LIST-ITEM-PAIRS writeable attribute.
    *
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItemPairs()
   {
      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];
      
      StringBuffer sb = new StringBuffer();
      char delimiter = config.delimiter;
      
      for (int i = 0; i < items.length; i++)
      {
         sb.append(items[i].getLabel().toStringMessage());
         sb.append(delimiter);
         sb.append(items[i].getValue().toStringMessage());
         if (i < items.length - 1)
            sb.append(delimiter);
      }
      
      return new character(sb.toString());
   }
   
   /**
    * Tell widget that data format should be honored.
    * 
    * @param    b
    *           If <code>true</code>, the format string will be honored.
    */
   public void setHonorFormat(boolean b)
   {
      config.honorFormat = b;
      pushScreenDefinition();
   }
   
   /**
    * Convert the given array to a list that is modifiable.  Note that the
    * <code>Arrays.asList()</code> method does NOT return a modifiable list.
    *
    * @param    items
    *           The array to be converted.
    *
    * @return   The modifiable list in the same order as the original array.
    */
   private List<ControlSetItem> asList(ControlSetItem[] items)
   {
      ArrayList<ControlSetItem> list = new ArrayList<ControlSetItem>();
      
      for (ControlSetItem item : items)
      {
         list.add(item);
      }
      
      return list;
   }
   
   /**
    * Sort items according to current sort flag. 
    */
   private void sortItems()
   {
      if (!config.sort || config.items == null)
         return;
      
      Arrays.sort(config.items, new ItemComparator());
   }

   /**
    * Local comparator for two Frame instances.
    */
   private static class ItemComparator
   implements Comparator<ControlSetItem>
   {
      /**
       * Compare two instances of <code>ControlSetItem</code> for label.
       * 
       * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
       */
      @Override
      public int compare(ControlSetItem o1, ControlSetItem o2)
      {
         return o1.getLabel().compareTo(o2.getLabel());
      }
   }

   /**
    * Helper interface (is standard in Java 8/Google Guava).
    */
   private static interface Function<A,V> {
      public V apply(A value);
   }

}
