/*
** Module   : RadioSetWidget.java
** Abstract : server side radioset widget implementation
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23494 Created initial version.
** 002 SIY 20050209   @24428 Major refactoring.
** 003 GES 20060803   @28382 Some cleanup and refactoring.
** 004 GES 20080320   @37593 Added TYPE attribute.
** 018 SIY 20111101          Updated imports.
** 019 CA  20130221          Refactored usage of TYPE attribute.
** 020 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 021 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 022 CA  20131115          Changes related to properly instantiate dynamic resources.
** 023 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 025 IAS 20150123          Implemented DELETE() method
** 026 IAS 20150202          Implemented ADD_LAST() method, fixed ADD-FIRST()
*/ 

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.ui.ControlSetEntity.*;
import com.goldencode.p2j.util.*;

/**
 * Server side radioset widget implementation.
 */
@LegacyResource(resource = LegacyResource.RADIO_SET)
public class RadioSetWidget
extends ControlSetEntity<RadioSetConfig>
{
   /**
    * Default constructor.
    */
   public RadioSetWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public RadioSetWidget(boolean dynamic)
   {
      super(dynamic, new RadioSetConfig());
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   horizontal
    *          The new value of the HORIZONTAL attribute.
    */
   public void setHorizontal(boolean horizontal)
   {
      config.horizontal = horizontal;
      pushScreenDefinition();
   }

   /**
    * Resets the value of the HORIZONTAL attribute for radio-set.
    * 
    * @param   horizontal
    *          The new value of the HORIZONTAL attribute.
    */
   public void setHorizontal(logical horizontal)
   {
      setHorizontal(horizontal.booleanValue());
   }
   
   /**
    * Get RADIO-BUTTONS attribute.
    * 
    * @return  Current value of the attribute.
    */
   public character getRadioButtons()
   {
      return getListItemPairs();
   }
   
   /**
    * Set RADIO-BUTTONS attribute. The parameter contains string
    * representation of all radio button labels and values separated by
    * delimiter (comma is the default value).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,..., LabelN,valueN</code>
    * 
    * @param   c
    *          New value of the attribute.
    */
   public void setRadioButtons(character c)
   {
      setListItemPairs(c);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param item
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character item)
   {
      if (item == null || item.isUnknown())
      {
         invalidDeleteArg();
         return new logical(false);
      }
      return delete(item.toStringMessage());
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param item
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(String item)
   {
      List<String> items = new ArrayList<>();
      items.add(item);
      logical rc = delete(items, GET_LABEL, true);
      if (!rc.booleanValue())
      {
         invalidDeleteArg();
      }
      return rc;
   }

   /**
    * Report invalid argument for delete().
    * 
    */
   private void invalidDeleteArg()
   {
      ErrorManager.recordOrShowError(4065,
            "The radio button in the DELETE attribute is invalid for " + type(), false);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(int64 idx)
   {
      UnimplementedFeature.unsupported("DELETE() method is undefined");
      return new logical();
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the
    * given index.
    * 
    * @param idx
    *           The index of the item to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(long idx)
   {
      UnimplementedFeature.unsupported("DELETE() method is undefined");
      return new logical();
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, BaseDataType value)
   {
      ControlSetItem[] items = config.items;
      if (items == null)
         items = new ControlSetItem[0];
      boolean selectable = config.autoResize || (items.length < config.heightChars);
      FrameConfig frameConfig = frame.getFrameWidget().getConfig();
      int maxSize = (int)frameConfig.heightChars;
      if( frameConfig.box) {
         maxSize -= 2;
      }
      if( maxSize <= items.length) {
         ErrorManager.recordOrShowError(4474, String.format(
               "New height for %s %s is too large to fit in frame.", type(),
                  frame.getName(this)), false);
         selectable = false;
      }
      ControlSetItem[] nitems = Arrays.copyOf(items, items.length + 1);
      ControlSetItem item = controlSetItem(label, value);
      item.setSelectable(selectable);
      nitems[items.length] = item;
      setItems(nitems);
      return new logical(true);
   }
   

   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   Always reports warning 4052 and returns "unknown".
    */
   @Override
   public logical addFirst(character label, BaseDataType value)
   {
      ErrorManager.recordOrShowError(4052, String.format(
            "ADD-FIRST is not a queryable attribute for %s %s.", type(),
               frame.getName(this)), false);
      return new logical();
   }
}
