/*
** Module   : DefaultList.java
** Abstract : Default scrollable item list implementation.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added javadoc.
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Moved ChUI-specific requestFocus impl to the ChUI impl classes.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 MAG 20140929 Added ensureRowVisible with scrolling option.
** 006 HC  20141021 Implemented window move/position support.
** 007 IAS 20150202 Added index() method.
** 008 IAS 20150204 Fix behavior to be compatible with 4GL
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.model.*;

/**
 * Base class for scrollable list implementations.
 * 
 * @param <T> 
 */
public abstract class DefaultList<T, O extends OutputManager<?>>
extends AbstractWidget<O>
implements Scrollable
{
   /** Enable item deselection flag. */
   private boolean allowDeselect = true;

   /** Number of columns in the list. */
   private double columns = 10;

   /** Current row in the list. */
   private int currentRow = 0;

   /** List model. */
   private ListModel<T> listModel;

   /** Storage for listeners. */
   private ArrayList<ScrollListener> scrollListeners = new ArrayList<ScrollListener>();

   /** List selection model. */
   private ListSelectionModel selectionModel = new ListSelectionModel();

   /** Number of visible rows in the list. */
   private int visibleRows = 5;

   /**
    * Constructor.
    * 
    * @param   model
    *          List model.
    */
   public DefaultList(ListModel<T> model)
   {
      listModel = model;
   }

   /**
    * Add scroll listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addScrollListener(ScrollListener listener)
   {
      scrollListeners.add(listener);
   }

   /**
    * Add selection listener.
    *  
    * @param   listener
    *          Listener to add.
    */
   public void addSelectionListener(SelectionListener listener)
   {
      selectionModel.add(listener);
   }

   /**
    * Control removing of selection from current item.
    * 
    * @param    allow
    *           <code>true</code> allow deselection of current item.
    */
   public void allowDeselect(boolean allow)
   {
      allowDeselect = allow;
   }

   /**
    * Clear list.
    */
   public void clear()
   {
      selectionModel.clear();
   }

   /**
    * Get current row in the list.
    * 
    * @return  current row.
    */
   public int currentRow()
   {
      return currentRow;
   }

   /**
    * Deselect specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void deselect(int from, int to)
   {
      selectionModel.deselect(from, to);
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Ensure that specified row is visible. Scroll list if necessary.
    * 
    * @param   row
    *          The row to process.
    */
   public void ensureRowVisible(int row)
   {
      ensureRowVisible(row, true);
   }

   /**
    * Ensure that specified row is visible. Scroll list if necessary.
    * 
    * @param   row
    *          The row to process.
    * @param   scroll
    *          Scrolling option.
    */
   public void ensureRowVisible(int row, boolean scroll)
   {
      if (!(parent() instanceof Viewport))
         return;

      currentRow = fixRow(row);

      if (scroll)
      {
         scroll(ScrollDirection.DOWN, 0);
         scroll(ScrollDirection.UP, row);
      }
   }

   /**
    * Get first selected row index.
    * 
    * @return  first selected row index.
    */
   public int first()
   {
      return selectionModel.first();
   }
   
   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return listModel.size();
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      columns = 1;

      for (int i = 0; i < listModel.size(); i++)
      {
         String c = listModel.get(i).toString();
         
         if (c.length() > columns)
            columns = c.length();
      }

      return new Dimension(columns, visibleRows);
   }

   /**
    * Get underlying list model.
    * 
    * @return  list model.
    */
   public ListModel<T> model()
   {
      return listModel;
   }

   /**
    * Scrollable widget preferred size.
    * 
    * @return  preferred size.
    */
   @Override
   public Dimension preferredSize()
   {
      return new Dimension(width(), visibleRows());
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
         return;

      int key = event.keyCode();

      if (focusChange(key))
         return;
      
      processKey(key);

      redraw();
   }

   /**
    * Process key event.
    * 
    * @param key
    *          key to process
    */
   protected void processKey(int key)
   {
      switch (key)
      {
         case Key.VK_CURSOR_DOWN :
            scroll(1);
            break;

         case Key.VK_PAGE_DOWN :
            scroll(visibleRows);
            break;

         case Key.VK_END :
            scroll(listModel.size());
            break;

         case Key.VK_CURSOR_UP :
            scroll(-1);
            break;

         case Key.VK_PAGE_UP :
            scroll(-visibleRows);
            break;

         case Key.VK_HOME :
            scroll(-listModel.size());
            break;
            
         case Key.VK_ENTER :
            selectItem();
            break;
      }
   }

   /**
    * Process scroll event.
    * 
    * @param   event
    *          Scroll event.
    */
   @Override
   public void processScrollEvent(ScrollEvent event)
   {
      for (ScrollListener listener : scrollListeners)
         listener.scroll(event);
   }

   /**
    * Select row with specified index.
    * 
    * @param   index
    *          Item index.
    */
   public void select(int index)
   {
      selectionModel.set(index, index);
   }

   /**
    * Select specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   public void select(int from, int to)
   {
      selectionModel.select(from, to);
   }

   /**
    * Check if item with specified index is selected.
    *  
    * @param   index
    *          Item index.
    *          
    * @return  <code>true</code> if item is selected.
    */
   public boolean selected(int index)
   {
      return selectionModel.selected(index);
   }

   /**
    * Get indices of all selected items.
    * 
    * @return  array of selected indices.
    */
   public int[] selection()
   {
      return selectionModel.selection();
   }

   /**
    * Get access to selection model.
    * 
    * @return  selection model.
    */
   public int selectionMode()
   {
      return selectionModel.mode();
   }

   /**
    * Set visible list width.
    * 
    * @param   cols
    *          Width in columns.
    */
   public void setColumns(double cols)
   {
      columns = Coordinate.scale(cols);
   }

   /**
    * Set list selection mode.
    * 
    * @param   mode
    *          Selection mode to set.
    */
   public void setSelectionMode(int mode)
   {
      selectionModel.setMode(mode);
   }

   /**
    * Set number of visible rows.
    * 
    * @param   rows
    *          Number of rows to set.
    */
   public void setVisibleRows(int rows)
   {
      visibleRows = rows;
   }

   /**
    * Get value of first selected item.
    * 
    * @return  value of first selected item.
    */
   public T value()
   {
      int index = selectionModel.first();

      if (index == -1)
         return null;

      return listModel.get(index);
   }

   /**
    * Get index of first selected item.
    * 
    * @return index of first selected item.
    */
   public int index()
   {
      return selectionModel.first();
   }

   /**
    * Get number of visible rows.
    * 
    * @return  number of visible rows.
    */
   public int visibleRows()
   {
      return visibleRows;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return columns;
   }

   /**
    * Fix row to make sure it fits into model bounds.
    * 
    * @param   row
    *          Row to fix.
    *          
    * @return  fixed row.
    */
   private int fixRow(int row)
   {
      if (row < 0 || listModel.size() < 1)
      {
         row = 0;
      }
      else if (row > listModel.size() - 1)
      {
         row = listModel.size() - 1;
      }
      return row;
   }

   /**
    * Common scrolling worker.
    *  
    * @param   delta
    *          Scrolling delta.
    */
   protected void scroll(int delta)
   {
      int save = currentRow;
      
      currentRow = fixRow(currentRow + delta);

      if (save > currentRow)
         scroll(ScrollDirection.DOWN, currentRow);
      else if (save < currentRow)
         scroll(ScrollDirection.UP, currentRow);
   }
   
   /**
    * Post scroll event with specified direction and row.
    * 
    * @param   dir
    *          Scrolling direction.
    * @param   row
    *          Scroll row.
    */
   private void scroll(ScrollDirection dir, int row)
   {
      EventManager.postEvent(new ScrollEvent(this, dir, new Point(0, row)));
   }

   /**
    * Handle selection/deselection of current item.
    */
   protected void selectItem()
   {
      if (listModel.size() > 0)
      {
         if (selected(currentRow))
         {
            if (allowDeselect)
               deselect(currentRow, currentRow);
            else
               selectionModel.notifyListeners();
         }
         else
         {
            if (selectionMode() == ListSelectionModel.SINGLE_SELECTION)
               select(currentRow);
            else
               select(currentRow, currentRow);
         }
         repaint();
      }
   }
   
   /**
    * Handle keys which may trigger focus change.
    * 
    * @param   key
    *          Key to check.
    *          
    * @return  <code>true</code> if key was processed.
    */
   private boolean focusChange(int key)
   {
      if (key == Key.VK_TAB)
      {
         parent().nextFocus();
         return true;
      }

      if (key == Key.VK_BACK_TAB)
      {
         parent().prevFocus();
         return true;
      }

       return false;
   }

   /**
    * Redrawing worker.
    */
   protected void redraw()
   {
      if (parent() instanceof Viewport)
         return;
      
      draw();
      requestFocus();
      requestSync();
   }
}
